"""Implements the functions used to search a directory for the files to be parsed by doXtrings"""
import logging
from os import walk
from pathlib import Path
from typing import Generator, List

from doxtrings.config import load_config

logger = logging.getLogger(__name__)


def search_files() -> Generator[Path, None, None]:
    """Runs through the directory specified in the config and yields file paths to be parsed.

    Yields:
        Generator[Path, None, None]: the file paths to be parsed
    """
    config = load_config()
    for dirpath, subdirs, filenames in walk(config.root_dir):
        logger.debug(f"Searching files at {dirpath}")
        for filename in filenames:
            if _should_include_file(
                filename, config.exclude_files, config.include_file_extensions
            ):
                file_path = Path(dirpath) / filename
                logger.debug(f"Yielding file {str(file_path)}")
                yield file_path

        # Updates subdir in place to remove unwanted subdirectories
        subdirs[:] = [
            subdir
            for subdir in subdirs
            if _should_include_dir(subdir, config.exclude_files)
        ]
        logger.debug(f"Looking into subdirs {subdirs}")


def _should_include_file(
    filename: str, exclude_list: List[str], allowed_extensions: List[str]
):
    suffix_match = _match_suffix(filename, allowed_extensions)
    if not suffix_match:
        logger.info(
            f"File {filename} excluded because extension does not match any of: {allowed_extensions}"
        )
    excluded = filename in exclude_list
    if excluded:
        logger.info(
            f"File {filename} excluded because found match in exclude list: {exclude_list}"
        )
    return suffix_match and not excluded


def _should_include_dir(dirname: str, exclude_list: List[str]):
    excluded = dirname in [excluded.strip("/") for excluded in exclude_list]
    if excluded:
        logger.info(
            f"Directory {dirname} excluded because found match in exclude list: {exclude_list}"
        )
    return not excluded


def _match_suffix(filename: str, allowed_suffixes: List[str]):
    return filename.endswith(tuple(allowed_suffixes))



"""The main CLI module, responsible for running all steps of doXtrings scan"""
from argparse import ArgumentParser
from typing import Optional

from doxtrings.config import set_config
from doxtrings.core.parser.file_parser import DoXtringsAstParser
from doxtrings.core.report import SimpleReporter
from doxtrings.core.search import search_files


def scan():
    """Runs the full doxtrings scan"""
    arg_parser = ArgumentParser(
        prog="DoXtrings", description="Detect missing and wrong docstrings"
    )

    arg_parser.add_argument("-c", "--config-file", default=None, required=False)
    args = arg_parser.parse_args()
    _scan(args.config_file)


def _scan(config_file: Optional[str]):
    set_config(config_file=config_file)
    parser = DoXtringsAstParser()
    reporter = SimpleReporter()
    for file in search_files():
        for documentable, diffs in parser.parse_file(file):
            reporter.add_documentable_diff(documentable=documentable, diffs=diffs)
    reporter.report()


if __name__ == "__main__":
    scan()



"""
Defines the doXtrings configuration object and setter/getter functions
"""
from __future__ import annotations

import json
import logging
from dataclasses import dataclass, field, fields, replace
from pathlib import Path
from types import ModuleType
from typing import Any, Dict, List, Optional, Tuple, Union

import tomli

from doxtrings.documentable import ALL_DOCUMENTABLE_TYPES, DocumentableType

_DEFAULT_CONFIG_FILES = (
    "pyproject.toml",
    "doxtrings.toml",
    ".doxtrings.toml",
    "doxtrings.json",
    ".doxtrings.json",
)

logger = logging.getLogger(__name__)


@dataclass
class IgnoreFilters:
    """
    Stores the ignore filters configuration.
    """

    ignore_prefixes_rules: Optional[List[str]] = None
    """
    A list of prefixes to Documentable's names that will mark them as ignored in the report.
    DEfaults to empty list.
    """
    include_prefixes_rules: Optional[List[str]] = None
    """
    A list of Documentable's name prefixes that will NOT be ignored in the report.
    Defaults to an empty list.
    """
    ignore_matches: Optional[List[str]] = None
    """
    A list of Documentable's names that will be ignored in the report.
    Defaults to empty list.
    """
    ignore_typing: Optional[bool] = None
    """
    If set to True, will not require typing information in the docstring.
    Defaults to None. Prefer using `get_ignore_typing` to get the value of this attribute,
    as it will account for `None` values.
    """
    ignore_arguments: Optional[bool] = None
    """
    If set to True, will not require arguments information in the docstring.
    Defaults to None. Prefer using `get_ignore_arguments` to get the value of this attribute,
    as it will account for `None` values.
    """
    ignore_return: Optional[bool] = None
    """
    If set to True, will not require return type information in the docstring.
    Defaults to None. Prefer using `get_ignore_arguments` to get the value of this attribute,
    as it will account for `None` values.
    """
    ignore_args_and_kwargs: Optional[bool] = None
    """
    If set to True, will not require *args and **kwargs in the docstring.
    Defaults to None. Prefer using `get_ignore_args_and_kwargs` to get the value of this attribute,
    as it will account for `None` values.
    """

    def get_ignore_typing(self) -> bool:
        """Gets ignore_typing value. Defaults to False if ignore_typing is None.

        Returns:
            bool: the ignore_typing property or False if not set
        """
        return False if self.ignore_typing is None else self.ignore_typing

    def get_ignore_arguments(self) -> bool:
        """Gets ignore_arguments value. Defaults to False if ignore_arguments is None.

        Returns:
            bool: the ignore_arguments property or False if not set
        """
        return False if self.ignore_arguments is None else self.ignore_arguments

    def get_ignore_return(self) -> bool:
        """Gets ignore_return value. Defaults to False if ignore_return is None.

        Returns:
            bool: the ignore_return property or False if not set
        """
        return False if self.ignore_return is None else self.ignore_return

    def get_ignore_args_and_kwargs(self) -> bool:
        """Gets ignore_args_and_kwargs value. Defaults to False if ignore_args_and_kwargs is None.

        Returns:
            bool: the ignore_args_and_kwargs property or False if not set
        """
        return (
            False
            if self.ignore_args_and_kwargs is None
            else self.ignore_args_and_kwargs
        )

    @classmethod
    def merged(cls, base: IgnoreFilters, *mergeables: IgnoreFilters) -> IgnoreFilters:
        """
        Returns a merged version of the provided IgnoreFilters.
        Merge happens from left to right, meaning values in the leftmost filters will be
        overwritten by the rightmost ones.

        Args:
            base (IgnoreFilters): the base filter to be merged into the others.
            *mergeables (IgnoreFilters): the other filters to be merged.

        Returns:
            IgnoreFilters: the merged version of the filters.
        """
        merged = replace(base)
        for a in mergeables:
            if merged.ignore_prefixes_rules is None:
                merged.ignore_prefixes_rules = a.ignore_prefixes_rules
            if merged.include_prefixes_rules is None:
                merged.include_prefixes_rules = a.include_prefixes_rules
            if merged.ignore_matches is None:
                merged.ignore_matches = a.ignore_matches
            if merged.ignore_typing is None:
                merged.ignore_typing = a.ignore_typing
            if merged.ignore_arguments is None:
                merged.ignore_arguments = a.ignore_arguments
            if merged.ignore_return is None:
                merged.ignore_return = a.ignore_return
            if merged.ignore_args_and_kwargs is None:
                merged.ignore_args_and_kwargs = a.ignore_args_and_kwargs
        return merged


@dataclass
class IgnoreFiltersRules:
    """
    A series of IgnoreFilters to be applied to specific types.
    """

    default: IgnoreFilters = field(default_factory=lambda: IgnoreFilters())
    """Ignore Filters for all documentables"""
    callables: IgnoreFilters = field(default_factory=lambda: IgnoreFilters())
    """Ignore Filters for all callables (functions and methods)"""
    functions: IgnoreFilters = field(default_factory=lambda: IgnoreFilters())
    """Ignore Filters for all functions"""
    methods: IgnoreFilters = field(default_factory=lambda: IgnoreFilters())
    """Ignore Filters for all methods"""
    classes: IgnoreFilters = field(default_factory=lambda: IgnoreFilters())
    """Ignore Filters for classes"""
    modules: IgnoreFilters = field(default_factory=lambda: IgnoreFilters())
    """Ignore Filters for modules"""
    assignments: IgnoreFilters = field(default_factory=lambda: IgnoreFilters())
    """Ignore Filters for all assignments (attributes, constants and variables)"""
    attributes: IgnoreFilters = field(default_factory=lambda: IgnoreFilters())
    """Ignore Filters for attributes"""
    constants: IgnoreFilters = field(default_factory=lambda: IgnoreFilters())
    """Ignore Filters for constants"""
    variables: IgnoreFilters = field(default_factory=lambda: IgnoreFilters())
    """Ignore Filters for variables"""


# TODO refactor config naming conventions
# TODO make recursive config
@dataclass
class DoXtringsConfig:
    """
    Class that stores all DoXtrings configuration.
    """

    root_dir: str = "./"
    """
    The root directory from where doxtrings will run the scan. Defaults to the current directory.
    """
    exclude_files: List[str] = field(
        default_factory=lambda: ["env", "venv", ".env", ".venv", "test/"]
    )
    """
    A list of files or directories to be excluded from the scan.
    Defaults to `["env", "venv", ".env", ".venv", "test/"]`
    """
    include_file_extensions: List[str] = field(default_factory=lambda: [".py"])
    """
    A list of file extensions that will be scanned.
    Defaults to `[".py"]`
    """
    types_to_check: List[str] = field(
        default_factory=lambda: list(ALL_DOCUMENTABLE_TYPES)
    )
    """
    A list with the types of Documentables that will be reported. Defaults to the list of all
    the available types (doxtrings.documentable:ALL_DOCUMENTABLE_TYPES)
    """
    docstring_format: str = "google"
    """
    The docstring format to use when parsing the docstring data. Possible values are "google",
    "epydoc", "rest" or "numpydoc". Defaults to "google".
    """

    parsing_depth: int = 2
    """
    How deep into the AST doxtrings will scan. Default value is 2.
    Using value of depth 2 means it will scan the root of a module and the first level of nesting,
    allowing for methods and attributes in classes to be scanned.
    """
    fail_fast: bool = False
    """If set to `True`, the scan will stop at the first difference found. Defaults to `False`"""

    fail_ignored_if_incorrect: bool = True
    """
    If set to `True`, documentables marked as ignored will still be added to the report if their
    docstring is not missing but is incorrect. Ignored types will also be added if incorrect.
    Defaults to `True`.
    """

    ignore_filters: IgnoreFiltersRules = field(
        default_factory=lambda: IgnoreFiltersRules()
    )
    """
    A dictionary of ignore rules specific for each type of documentable where the keys are the
    documentable types and the values are the specific rules. If any rule is specified
    it will override the default ignore rules for that type of documentable.
    """

    def get_ignore_filters(self, type: Union[DocumentableType, str]) -> IgnoreFilters:
        """Gets the ignore filters based on the type of documentable

        Args:
            type (Union[DocumentableType, str]): the type of documentable

        Raises:
            ValueError: If `type` is not a valid documentable type

        Returns:
            IgnoreFilters: Returns the ignore filters for the specific type
        """
        if type not in ALL_DOCUMENTABLE_TYPES:
            raise ValueError(
                f"Cannot get ignore filter for {type}. Valid types are {ALL_DOCUMENTABLE_TYPES}"
            )
        type_str = type.value if isinstance(type, DocumentableType) else type
        if type_str == DocumentableType.FUNCTION.value:
            return IgnoreFilters.merged(
                self.ignore_filters.functions,
                self.ignore_filters.callables,
                self.ignore_filters.default,
            )
        elif type_str == DocumentableType.METHOD.value:
            return IgnoreFilters.merged(
                self.ignore_filters.methods,
                self.ignore_filters.callables,
                self.ignore_filters.default,
            )
        elif type_str == DocumentableType.CLASS.value:
            return IgnoreFilters.merged(
                self.ignore_filters.classes, self.ignore_filters.default
            )
        elif type_str == DocumentableType.MODULE.value:
            return IgnoreFilters.merged(
                self.ignore_filters.modules, self.ignore_filters.default
            )
        elif type_str == DocumentableType.ATTRIBUTE.value:
            return IgnoreFilters.merged(
                self.ignore_filters.attributes,
                self.ignore_filters.assignments,
                self.ignore_filters.default,
            )
        elif type_str == DocumentableType.VARIABLE.value:
            return IgnoreFilters.merged(
                self.ignore_filters.variables,
                self.ignore_filters.assignments,
                self.ignore_filters.default,
            )
        elif type_str == DocumentableType.CONSTANT.value:
            return IgnoreFilters.merged(
                self.ignore_filters.constants,
                self.ignore_filters.assignments,
                self.ignore_filters.default,
            )
        else:
            raise NotImplementedError(
                f"Config does not have support for {type}, even though type is valid"
            )


_ROOT_CONFIG_FIELDS = tuple(f.name for f in fields(DoXtringsConfig))
_IGNORE_FILTERS_FIELDS = tuple(f.name for f in fields(IgnoreFilters))
_IGNORE_FILTERS_BY_TYPE_FIELDS = tuple(f.name for f in fields(IgnoreFiltersRules))


class InvalidConfigFileTypeError(Exception):
    """Exception raised if configuration file is in invalid format"""


_config = None


def set_config(config_file: Optional[str] = None):
    """Sets the global DoXtrings configuration object.

    Args:
        config_file (Optional[str], optional): the file from which to parse the configuration.
            Defaults to None.
    """
    config = _initialize_config(config_file)
    global _config
    _config = config


def load_config() -> DoXtringsConfig:
    """Loads the global configuration

    Raises:
        ValueError: raised if config was not previously set using 'set_config'

    Returns:
        DoXtringsConfig: returns the DoxtringsConfig object.
    """
    if _config is None:
        raise ValueError("Config loaded before it was initialized")
    return _config


def _initialize_config(config_file: Optional[str] = None) -> DoXtringsConfig:
    if config_file:
        if not Path(config_file).is_file():
            raise FileNotFoundError(f"File {config_file} does not exist")
    config_files = (
        [Path(config_file)]
        if config_file is not None
        else [Path(f) for f in _DEFAULT_CONFIG_FILES]
    )
    for file_path in config_files:
        logger.info(f"Looking for config file at {str(file_path)}")
        cfg_dict = _load_cfg_if_exists(file_path)
        if cfg_dict is not None:
            logger.info(f"Read config file {str(file_path)}")
            return _build_config(cfg_dict)
    return DoXtringsConfig()


def _build_config(cfg_dict: Dict[str, Any]) -> DoXtringsConfig:
    _check_fields(cfg_dict, valid_fields=_ROOT_CONFIG_FIELDS)
    ignore_filters_rules_dict = cfg_dict.pop("ignore_filters", {})
    _check_fields(
        ignore_filters_rules_dict,
        valid_fields=_IGNORE_FILTERS_BY_TYPE_FIELDS,
        field_prefix="ignore_filters.",
    )
    ignore_filters_rules_kwargs: Dict[str, IgnoreFilters] = {}
    for field in _IGNORE_FILTERS_BY_TYPE_FIELDS:
        ignore_dict: Dict[str, Any] = ignore_filters_rules_dict.get(field, {})
        _check_fields(
            ignore_dict,
            valid_fields=_IGNORE_FILTERS_FIELDS,
            field_prefix=f"ignore_filters.{field}.",
        )
        ignore_filters_rules_kwargs[field] = IgnoreFilters(**ignore_dict)
    ignore_filters_rules = IgnoreFiltersRules(**ignore_filters_rules_kwargs)

    return DoXtringsConfig(
        **cfg_dict,
        ignore_filters=ignore_filters_rules,
    )


def _check_fields(
    cfg_dict: Dict[str, Any], valid_fields: Tuple[str, ...], field_prefix: str = ""
):
    invalid_fields: List[str] = []
    for key in cfg_dict.keys():
        if key not in valid_fields:
            invalid_fields.append(field_prefix + key)
    if len(invalid_fields) > 0:
        raise TypeError(
            f"Unexpected DoXtringsConfig fields: {', '.join(sorted(invalid_fields))}"
        )


def _load_cfg_if_exists(cfg_path: Path) -> Optional[Dict[str, Any]]:
    if cfg_path.exists() and cfg_path.is_file():
        parsed_dict = _parse_file(cfg_path)
        # If file is pyproject.toml, get only the correct section
        if cfg_path.name == "pyproject.toml":
            logger.info(
                f"Config file is pyproject.toml, will look for specific section"
            )
            toml_dict = _get_pyproject_toml_section(parsed_dict)
            if toml_dict is None:
                logger.info(f"doxtrings config not found in pyproject.toml")
                return None
            else:
                parsed_dict = toml_dict
        return parsed_dict
    else:
        return None


def _parse_file(cfg_path: Path) -> Dict[str, Any]:
    parser_module = _get_parser_module(cfg_path)
    with cfg_path.open("rb") as cfg_file:
        parsed_dict: Dict[str, Any] = parser_module.load(cfg_file)

        return parsed_dict


def _get_parser_module(cfg_file: Path) -> ModuleType:
    if cfg_file.suffix == ".toml":
        return tomli
    elif cfg_file.suffix == ".json":
        return json
    else:
        raise InvalidConfigFileTypeError(
            f"Allowed config types are '.toml' and '.json', but found '{cfg_file.suffix}'"
        )


def _get_pyproject_toml_section(
    parsed_dict: Dict[str, Any]
) -> Optional[Dict[str, Any]]:
    try:
        return parsed_dict["tool"]["doxtrings"]
    except:
        return None



""" Implements the class responsible for parsing each file into Documentables"""
import ast
import logging
from pathlib import Path
from typing import Dict, Generator, List, Optional, Tuple, Type

from doxtrings.config import DoXtringsConfig, load_config
from doxtrings.core.parser.node_parser import ParsingContext
from doxtrings.documentable import Diff, DiffReasons, Documentable

logger = logging.getLogger(__name__)

from doxtrings.core.parser.node_parser import (
    BaseNodeParser,
    ClassParser,
    FunctionParser,
    ModuleParser,
    VariablesParser,
)

_NodeParserMapping = Dict[Type[ast.AST], List[BaseNodeParser]]
"""Type alias to a mapping of node types to parsers"""


class DoXtringsAstParser:
    """Parser class that reads a file and returns a Generator of Documentables"""

    def __init__(self):
        """Initializes a new DoXtringsAstParser"""
        self._config = load_config()
        self._node_parser_mapping = _get_node_to_parser_mapping(self._config)

    def parse_file(
        self, file: Path
    ) -> Generator[Tuple[Documentable, List[Diff]], None, None]:
        """Parses a file into a list of documentables

        Args:
            file (Path): path of the file to be parsed

        Yields:
            Generator[Tuple[Documentable, List[Diff]], None, None]: generator of tuples
            with documentables and an optional namedtuple with the docstring data.
        """
        logger.info(f"Parsing file {str(file)}")
        self._filename = file.name
        self._filepath = file
        module = _ast_parse(file)
        return self._parse_ast_node(module)

    def _parse_ast_node(
        self,
        node: ast.AST,
        context: List[Type[ast.AST]] = [],
        next_node: Optional[ast.AST] = None,
        ignored_node: bool = False,
    ) -> Generator[Tuple[Documentable, List[Diff]], None, None]:
        parsers = self._node_parser_mapping.get(type(node), [])
        logger.debug(f"Parsing node {node} with {parsers}")
        for parser in parsers:
            node_data = parser.parse_node(
                node,
                ParsingContext(
                    self._filepath, self._filename, context, next_node, ignored_node
                ),
            )
            # If node is ignored, we mark all the next nodes parsed inside this one as ignored
            if node_data and self._is_documentable_ignored(node_data):
                ignored_node = True

            if node_data and node_data.documentable_type in self._config.types_to_check:
                diffs = parser.get_diffs()
                logger.debug(f"Found data for node {node}")
                if len(diffs) and self._should_report_diffs(diffs, ignored_node):
                    logger.info(f"Found diffs for node {node}, yielding")
                    yield node_data, diffs

        # Traverse all child nodes
        if len(context) < self._config.parsing_depth:
            for result in self._traverse_ast(
                node, context=context, ignored_node=ignored_node
            ):
                yield result

    def _traverse_ast(
        self, node: ast.AST, context: List[Type[ast.AST]], ignored_node: bool = False
    ) -> Generator[Tuple[Documentable, List[Diff]], None, None]:
        new_context = context + [type(node)]
        for _, value in ast.iter_fields(node):
            if isinstance(value, List):
                # We know nothing about type of item, so we check if item is AST
                # and ignore the type errors pylance complains below
                for i, item in enumerate(value):  # type:ignore
                    if isinstance(item, ast.AST):
                        next_node: Optional[ast.AST] = None
                        if i + 1 < len(value) and isinstance(  # type:ignore
                            value[i + 1], ast.AST
                        ):
                            next_node = value[i + 1]  # type:ignore
                        for result in self._parse_ast_node(
                            item, new_context, next_node, ignored_node  # type:ignore
                        ):
                            yield result

            elif isinstance(value, ast.AST):
                for result in self._parse_ast_node(
                    value, new_context, ignored_node=ignored_node
                ):
                    yield result

    def _is_documentable_ignored(self, documentable: Documentable) -> bool:
        ignore_rules = self._config.get_ignore_filters(documentable.documentable_type)
        ignore_matches = (
            [] if ignore_rules.ignore_matches is None else ignore_rules.ignore_matches
        )
        ignore_prefixes_rules = (
            []
            if ignore_rules.ignore_prefixes_rules is None
            else ignore_rules.ignore_prefixes_rules
        )
        include_prefixes_rules = (
            []
            if ignore_rules.include_prefixes_rules is None
            else ignore_rules.include_prefixes_rules
        )
        return any([documentable.name == ignored for ignored in ignore_matches]) or (
            documentable.name.startswith(tuple(ignore_prefixes_rules))
            and not documentable.name.startswith(tuple(include_prefixes_rules))
        )

    def _should_report_diffs(self, diffs: List[Diff], is_documentable_ignored: bool):
        # If documentable is ignored and config is set
        # not to fail if ignored are incorrect, we do not add it to report in any circumstance
        if is_documentable_ignored and self._config.fail_ignored_if_incorrect is False:
            return False

        # If the documentable is ignored and the only diff found for it is that the docstring
        # is missing, then we should ignore it as well
        if (
            is_documentable_ignored
            and len(diffs) == 1
            and diffs[0].reason == DiffReasons.NO_DOCSTRING
        ):
            return False

        # For all other cases, we do want to yield the diffs
        return True


def _get_node_to_parser_mapping(config: DoXtringsConfig) -> _NodeParserMapping:
    mapping: _NodeParserMapping = {}
    parsers = _load_node_parsers()
    for parser in parsers:
        _add_inputs_in_mapping(mapping, parser)
    return mapping


def _add_inputs_in_mapping(
    mapping: _NodeParserMapping,
    parser: BaseNodeParser,
):
    inputs = parser.get_input_types()
    for input in inputs:
        if input not in mapping:
            mapping[input] = []
        mapping[input].append(parser)


def _load_node_parsers() -> Tuple[BaseNodeParser, ...]:
    # TODO make this extensible
    return (FunctionParser(), ClassParser(), ModuleParser(), VariablesParser())


def _ast_parse(file: Path) -> ast.Module:
    file_content: str
    with file.open("r") as file_input:
        file_content = file_input.read()
    return ast.parse(file_content, file.name)



"""
Implements the classes used to print the reports of the found differences
found in the source code
"""
import sys
from abc import ABC, abstractmethod
from typing import List, NamedTuple, Union

from doxtrings.config import load_config
from doxtrings.core.parser.node_parser import Diff
from doxtrings.documentable import Documentable


class ReportItem(NamedTuple):
    """A NamedTuple containing the relevant report data for a single documentable"""

    documentable: Documentable
    """The documentable this report item is referring to"""
    diffs: List[Diff]
    """The documentable's list of differences"""


class BaseReporter(ABC):
    """Base reporter class"""

    def __init__(self):
        """Initializes the base reporter with an empty list of report items"""
        self._config = load_config()
        self.report_items: List[ReportItem] = []

    def add_documentable_diff(self, documentable: Documentable, diffs: List[Diff]):
        """
        Adds a list of diffs to a documentable. The provided `diffs` argument is filtered
        according to configuration rules before being linked to the documentable in a ReportItem.

        Args:
            documentable (Documentable): the documentable to be added to the report
            diffs (List[Diff]): all the diffs found for the documentable
        """
        if len(diffs):
            self.report_items.append(ReportItem(documentable, diffs))
            if self._config.fail_fast:
                self.report()

    @abstractmethod
    def report(self):
        """Abstract method responsible for printing out the report and handling exit codes"""


class SimpleReporter(BaseReporter):
    """A simple reporter that prints out the relevant information"""

    def report(self):
        """
        Prints out all the collected differences in the console.
        If there are diffs to be reported, exits with code 1.
        """
        if len(self.report_items) == 0:
            print("All docstrings have passed the check")
            sys.exit(0)
        else:
            for item in self.report_items:
                file_line_ref = f"{item.documentable.file}:{item.documentable.line}"
                print(
                    f"Docstring error found at {file_line_ref} on {item.documentable.documentable_type} {item.documentable.name}"
                )
                for diff in item.diffs:
                    location = get_location_string(diff.location)
                    line = f"    At {location}: Reason {diff.reason.value}"
                    if diff.code_value or diff.docstring_value:
                        line += f" expected {diff.code_value} but docstring is {diff.docstring_value}"
                    print(line)
            sys.exit(1)


def get_location_string(location: List[Union[str, int]]) -> str:
    """Transforms a location list into a human readable string.

    Args:
        location (List[Union[str, int]]): the list containing location information.

    Returns:
        str: a string representation of the location list.
    """
    if len(location) == 0:
        return "root"
    str_parts = [f".{l}" if isinstance(l, str) else f"[{str(l)}]" for l in location]
    return "".join(str_parts)



""" Implements classes capable of parsing both the AST nodes and the docstrings into Documentables"""
import ast
import logging
from abc import ABC, abstractmethod
from difflib import SequenceMatcher
from itertools import zip_longest
from pathlib import Path
from typing import Any, List, NamedTuple, Optional, Set, Tuple, Type, Union

import docstring_parser as dp

from doxtrings.config import IgnoreFilters, load_config
from doxtrings.documentable import Diff, DiffReasons, Documentable, DocumentableType

logger = logging.getLogger(__name__)


class EmptyTuple(NamedTuple):
    """An empty tuple used when there is no relevant data to be stored"""


class ParsingContext(NamedTuple):
    """
    This is a context regarding a node that is being parsed.
    Includes the filename and the parent nodes. This is useful to determine if a function is a
    method, for instance.
    """

    filepath: Path
    """The path to the file being parsed"""
    filename: str
    """The name of the file being parsed"""
    parent_node_types: List[Type[ast.AST]]
    """A list with the types of the parent nodes of the current node being parsed"""
    next_node: Optional[ast.AST]
    """The node next to this one, if there is any"""
    is_ignored: bool
    """Wether this node is marked as ignored"""


class BaseNodeParser(ABC):
    """Defines the base interface for a node parser"""

    _input: ast.AST
    _context: ParsingContext
    _name: str
    _documentable_type: Optional[DocumentableType]
    _lineno: int

    def __init__(self):
        """Initializes the parser"""
        self._config = load_config()

    @classmethod
    @abstractmethod
    def get_input_types(cls) -> Set[Type[ast.AST]]:
        """Returns a set with the AST types this parser can parse from

        Returns:
            Set[Type[ast.AST]]: the set of AST types this parser class can parse from
        """

    @classmethod
    @abstractmethod
    def get_output_types(cls) -> Set[DocumentableType]:
        """Returns a set with the DocumentableTypes this parser can parse to

        Returns:
            Set[DocumentableType]: a set with the DocumentableTypes this parser can parse to
        """

    def parse_node(
        self,
        input: ast.AST,
        context: ParsingContext,
    ) -> Optional[Documentable]:
        """Concrete method an AST node into a Documentable. Usually should not be overwritten by
        child classes. Child classes should instead implement `_parse_node`.

        Args:
            input (ast.AST): the AST node to be parsed
            context (ParsingContext): the parsing context for this node

        Returns:
            Optional[Documentable]: the parsed Documentable. Might be None if the parser could
            not parse the input.
        """
        self._input = input
        self._context = context
        self._name = self._get_node_name()
        self._documentable_type = self._get_documentable_type()
        self._lineno = input.lineno if hasattr(input, "lineno") else 0
        if self._documentable_type:
            self._code_data = self._parse_node()
            self._docstring_data = self._parse_docstring()
            return Documentable(
                file=self._context.filepath,
                line=self._lineno,
                name=self._name,
                documentable_type=self._documentable_type,
            )
        else:
            self._code_data = None
            self._docstring_data = None

    def get_diffs(self) -> List[Diff]:
        """Gets the list of diffs for the node

        Returns:
            List[Diff]: List of Diffs for the node
        """
        if self._docstring_data is None:
            return [
                Diff(DiffReasons.NO_DOCSTRING, location=[], code_value=self._code_data)
            ]
        elif self._docstring_data != self._code_data:
            return [
                Diff(
                    DiffReasons.CONFLICTING_VALUE,
                    location=[],
                    code_value=self._code_data,
                    docstring_value=self._docstring_data,
                )
            ]
        else:
            return []

    def _get_error_location_msg(self):
        return (
            f"Error found at {self._documentable_type} {self._name} at "
            f"{self._context.filepath}:{self._lineno}. "
        )

    def _get_node_name(self) -> str:
        """Gets the name of the input node.

        Returns:
            str: the name of the documentable
        """
        if hasattr(self._input, "name"):
            return str(self._input.name)  # type: ignore (we just checked for this attribute)
        else:
            raise NotImplementedError(
                self._get_error_location_msg()
                + f"`_get_node_name` not implemented for {type(self)} and default implementation failed"
            )

    def _get_documentable_type(self) -> Optional[DocumentableType]:
        """Determines the input node's type of the Documentable

        Returns:
            Optional[DocumentableType]: the documentable type.
                If None, this means this parser could not parse the node.
        """
        outputs = self.get_output_types()
        if len(outputs) == 1:
            return outputs.pop()
        else:
            raise NotImplementedError(
                self._get_error_location_msg()
                + f"`_get_documentable_type` not implemented for {self} and default implementation failed"
            )

    def _parse_node(
        self,
    ) -> NamedTuple:
        """Parses the AST node into a NamedTuple.

        Returns:
            NamedTuple: the parsed NamedTuple. Might be None if
            the parser could not parse the input.
        """
        return EmptyTuple()

    @abstractmethod
    def _parse_docstring(self) -> Optional[NamedTuple]:
        """Parses the node's docstring into a NamedTuple.

        Returns:
            Optional[NamedTuple]: the parsed NamedTuple. Might be None if there is no docstring.
        """


class ModuleParser(BaseNodeParser):
    """Parser for module nodes"""

    @classmethod
    def get_input_types(cls) -> Set[Type[ast.AST]]:
        """Returns a set with the AST types this parser can parse from.

        Returns:
            Set[Type[ast.AST]]: returns `set([ast.Module])`
        """
        return set([ast.Module])

    @classmethod
    def get_output_types(cls) -> Set[DocumentableType]:
        """Returns a set with the DocumentableTypes this parser can parse to

        Returns:
            Set[DocumentableType]: returns `set([DocumentableType.MODULE])`
        """
        return set([DocumentableType.MODULE])

    def _get_node_name(self) -> str:
        """
        Gets the name of the module. Because module nodes do not have a name attribute,
        we get the name from the context's filename

        Returns:
            str: the name of the documentable
        """
        return self._context.filename.rpartition(".")[0]

    def _parse_docstring(self) -> Optional[EmptyTuple]:
        """
        Parses the node's docstring into an EmptyTuple. There is no relevant data to be checked in
        the module docstring, so we just return the EmptyTuple if we find a docstring.

        Returns:
            Optional[EmptyTuple]: returns an EmptyTuple if there is a docstring for the module.
        """
        assert isinstance(self._input, ast.Module)
        docstring = ast.get_docstring(self._input)
        if docstring and len(docstring.strip()):
            return EmptyTuple()
        return None


class VariablesParser(BaseNodeParser):
    """Parser for attribute nodes"""

    @classmethod
    def get_input_types(cls) -> Set[Type[ast.AST]]:
        """Returns a set with the AST types this parser can parse from.

        Returns:
            Set[Type[ast.AST]]: returns `set([ast.ClassDef])`
        """
        return set([ast.Assign, ast.AnnAssign])

    @classmethod
    def get_output_types(cls) -> Set[DocumentableType]:
        """Returns a set with the DocumentableTypes this parser can parse to

        Returns:
            Set[DocumentableType]: returns `set([DocumentableType.CLASS])`
        """
        return set(
            [
                DocumentableType.CONSTANT,
                DocumentableType.ATTRIBUTE,
                DocumentableType.VARIABLE,
            ]
        )

    def _parse_docstring(self) -> Optional[EmptyTuple]:
        """
        Parses the node's docstring into an EmptyTuple. There is no relevant data to be checked in
        the variable's docstring, so we just return the EmptyTuple if we find a docstring.

        Returns:
            Optional[EmptyTuple]: returns an EmptyTuple if there is a docstring for the class.
        """
        assert isinstance(self._input, (ast.Assign, ast.AnnAssign))
        docstring = VariablesParser._get_variable_docstring(self._context.next_node)
        if docstring and len(docstring.strip()):
            return EmptyTuple()
        return None

    def _get_node_name(self) -> str:
        """Gets the name of the input node.

        Returns:
            str: the name of the documentable
        """
        opt_name = self._get_node_name_opt()
        return opt_name if opt_name else ""

    def _get_node_name_opt(self) -> Optional[str]:
        """Gets the optional name of the input node.

        Returns:
            Optional[str]: the name of the documentable or `None` if it is not a valid documentable
        """
        assert isinstance(self._input, (ast.Assign, ast.AnnAssign))
        ann_assign_name = VariablesParser._get_ann_assign_name(self._input)
        assign_name = VariablesParser._get_assign_name(self._input)
        return ann_assign_name if ann_assign_name else assign_name

    def _get_documentable_type(self) -> Optional[DocumentableType]:
        """Determines the input node's type of the Documentable

        Returns:
            Optional[DocumentableType]: the documentable type.
                If None, this means this parser could not parse the node.
        """
        name = self._get_node_name_opt()
        if not name:
            return None

        # Constants and Variables are only considered when in the root of the module
        if len(self._context.parent_node_types) == 1:
            # Here, what what differs a constant from a variable is only the case of the name
            if name == name.upper():
                return DocumentableType.CONSTANT
            else:
                return DocumentableType.VARIABLE
        elif self._context.parent_node_types[-1] == ast.ClassDef:
            return DocumentableType.ATTRIBUTE
        else:
            return None

    @staticmethod
    def _get_ann_assign_name(node: ast.AST) -> Optional[str]:
        if isinstance(node, ast.AnnAssign) and isinstance(node.target, ast.Name):
            return node.target.id
        else:
            return None

    @staticmethod
    def _get_assign_name(node: ast.AST) -> Optional[str]:
        if (
            isinstance(node, ast.Assign)
            and len(node.targets) == 1
            and isinstance(node.targets[0], ast.Name)
        ):
            return node.targets[0].id
        else:
            return None

    @staticmethod
    def _get_variable_docstring(next_node: Optional[ast.AST]):
        if (
            next_node
            and isinstance(next_node, ast.Expr)
            and isinstance(next_node.value, ast.Constant)
            and isinstance(next_node.value.value, str)
        ):
            return next_node.value.value
        else:
            return None


class ClassParser(BaseNodeParser):
    """Parser for class nodes"""

    @classmethod
    def get_input_types(cls) -> Set[Type[ast.AST]]:
        """Returns a set with the AST types this parser can parse from.

        Returns:
            Set[Type[ast.AST]]: returns `set([ast.ClassDef])`
        """
        return set([ast.ClassDef])

    @classmethod
    def get_output_types(cls) -> Set[DocumentableType]:
        """Returns a set with the DocumentableTypes this parser can parse to

        Returns:
            Set[DocumentableType]: returns `set([DocumentableType.CLASS])`
        """
        return set([DocumentableType.CLASS])

    def _parse_docstring(self) -> Optional[EmptyTuple]:
        """
        Parses the node's docstring into an EmptyTuple. There is no relevant data to be checked in
        the class docstring, so we just return the EmptyTuple if we find a docstring.

        Returns:
            Optional[EmptyTuple]: returns an EmptyTuple if there is a docstring for the class.
        """
        assert isinstance(self._input, ast.ClassDef)
        docstring = ast.get_docstring(self._input)
        if docstring and len(docstring.strip()):
            return EmptyTuple()
        return None


class FunctionArgument(NamedTuple):
    """relevant data of a function argument"""

    name: str
    """Name of the function argument"""
    type_hint: Optional[str]
    """Function argument type hint"""


class FunctionData(NamedTuple):
    """Relevant data for a function"""

    arguments: Tuple[FunctionArgument, ...]
    """The function arguments as a tuple of FunctionArgument"""
    return_type: Optional[str]
    """the return type of the function"""


class FunctionParser(BaseNodeParser):
    """Parser for function nodes"""

    _code_data: FunctionData
    _docstring_data: Optional[FunctionData]

    # TODO move this to config
    _DOCSTRING_STYLES = {
        "google": dp.DocstringStyle.GOOGLE,
        "epydoc": dp.DocstringStyle.EPYDOC,
        "rest": dp.DocstringStyle.REST,
        "numpydoc": dp.DocstringStyle.NUMPYDOC,
    }

    @classmethod
    def get_input_types(cls) -> Set[Type[ast.AST]]:
        """Returns a set with the AST types this parser can parse from.

        Returns:
            Set[Type[ast.AST]]: returns `set([ast.FunctionDef, ast.AsyncFunctionDef])`
        """
        return set([ast.FunctionDef, ast.AsyncFunctionDef])

    @classmethod
    def get_output_types(cls) -> Set[DocumentableType]:
        """Returns a set with the DocumentableTypes this parser can parse to

        Returns:
            Set[DocumentableType]: returns `set([DocumentableType.FUNCTION, DocumentableType.METHOD])`
        """
        return set([DocumentableType.FUNCTION, DocumentableType.METHOD])

    def _get_documentable_type(self) -> Optional[DocumentableType]:
        """Determines the input node's type of the Documentable

        Returns:
            Optional[DocumentableType]: the documentable type.
                If None, this means this parser could not parse the node.
        """
        return (
            DocumentableType.METHOD
            if self._context.parent_node_types[-1] == ast.ClassDef
            else DocumentableType.FUNCTION
        )

    def _parse_node(
        self,
    ) -> FunctionData:
        assert isinstance(self._input, (ast.FunctionDef, ast.AsyncFunctionDef))
        is_method = self._documentable_type == DocumentableType.METHOD
        ordered_args = (
            self._input.args.posonlyargs
            + self._input.args.args
            + FunctionParser._prepare_vararg(self._input.args.vararg, "*")
            + self._input.args.kwonlyargs
            + FunctionParser._prepare_vararg(self._input.args.kwarg, "**")
        )

        if is_method and len(ordered_args):
            # Removing self or cls from docstrings
            if ordered_args[0].arg == "self" or ordered_args[0].arg == "cls":
                ordered_args.pop(0)

        function_args = tuple(self._get_arg_data(arg) for arg in ordered_args)
        return_type = self._get_type_annotation(self._input.returns)
        return FunctionData(arguments=function_args, return_type=return_type)

    @staticmethod
    def _prepare_vararg(arg: Optional[ast.arg], prefix: str) -> List[ast.arg]:
        if arg is None:
            return []
        arg.arg = prefix + arg.arg
        return [arg]

    def _get_arg_data(self, arg: ast.arg) -> FunctionArgument:
        return FunctionArgument(
            name=arg.arg, type_hint=self._get_type_annotation(arg.annotation)
        )

    def _get_type_annotation(
        self, annotation_node: Optional[ast.expr]
    ) -> Optional[str]:
        type_hint: Optional[str] = None
        if annotation_node is None:
            return type_hint
        elif isinstance(annotation_node, ast.Name):
            type_hint = str(annotation_node.id)
        elif isinstance(annotation_node, ast.Constant):
            type_hint = self._get_constant_string_value(annotation_node.value)

        # Recursive methods
        elif isinstance(annotation_node, ast.Subscript):
            type_hint = self._get_subscript_type_hint(annotation_node)
        elif isinstance(annotation_node, ast.BinOp):
            type_hint = self._get_binop_type_hint(annotation_node)
        elif isinstance(annotation_node, ast.UnaryOp):
            type_hint = self._get_unop_type_hint(annotation_node)
        elif isinstance(annotation_node, ast.Tuple):
            type_hint = self._get_tuple_type_hint(annotation_node)
        elif isinstance(annotation_node, ast.List):
            type_hint = self._get_list_type_hint(annotation_node)
        elif isinstance(annotation_node, ast.Attribute):
            type_hint = self._get_attribute_type_hint(annotation_node)
        elif isinstance(annotation_node, ast.Index):
            # ast.Index was deprecated in python 3.9
            # Since we want to support python 3.8, we need to ignore typing for this special case
            type_hint = self._get_type_annotation(annotation_node.value)  # type: ignore
        else:
            raise NotImplementedError(
                self._get_error_location_msg()
                + f"Type {type(annotation_node)} is not a valid type annotation. "
                + "This means you should probably change the type annotation to a valid one. "
                + "If you think it is a valid type annotation raise an issue in doXtrings repo."
            )
        return type_hint

    def _get_constant_string_value(self, value: Any) -> str:
        if value is None or isinstance(value, (int, float, bool)):
            return str(value)
        if isinstance(value, str):
            return f'"{value}"'
        elif value == Ellipsis:
            return "..."
        else:
            raise NotImplementedError(
                self._get_error_location_msg()
                + f"Type hint parsing not implemented for constant type {value}"
            )

    def _get_subscript_type_hint(self, annotation_node: ast.Subscript):
        outer_value = self._get_type_annotation(annotation_node.value)
        inner_value = self._get_type_annotation(annotation_node.slice)

        return f"{outer_value}[{inner_value}]"

    def _get_unop_type_hint(self, annotation_node: ast.UnaryOp):
        operand = self._get_type_annotation(annotation_node.operand)
        if isinstance(annotation_node.op, ast.USub):
            operator = "-"
        elif isinstance(annotation_node.op, ast.UAdd):
            operator = "+"
        elif isinstance(annotation_node.op, ast.Not):
            operator = "not "
        elif isinstance(annotation_node.op, ast.Invert):
            operator = "~"
        else:
            raise ValueError(
                self._get_error_location_msg()
                + f"Unknown Unary operator type {type(annotation_node.op)}"
            )

        return f"{operator}{operand}"

    def _get_binop_type_hint(self, annotation_node: ast.BinOp):
        left_value = self._get_type_annotation(annotation_node.left)
        right_value = self._get_type_annotation(annotation_node.right)

        return f"{left_value} | {right_value}"

    def _get_tuple_type_hint(self, annotation_node: ast.Tuple) -> str:
        annotations = [self._get_type_annotation(elt) for elt in annotation_node.elts]
        return ", ".join(["None" if ann is None else ann for ann in annotations])

    def _get_list_type_hint(self, annotation_node: ast.List) -> str:
        annotations = [self._get_type_annotation(elt) for elt in annotation_node.elts]
        return (
            "["
            + ", ".join(["None" if ann is None else ann for ann in annotations])
            + "]"
        )

    def _get_attribute_type_hint(self, annotation_node: ast.Attribute):
        outer_value = self._get_type_annotation(annotation_node.value)

        return f"{outer_value}.{annotation_node.attr}"

    def _parse_docstring(self) -> Optional[FunctionData]:
        """
        Parses the node's docstring into a FunctionData tuple.

        Returns:
            Optional[FunctionData]: returns the parsed FunctionData if there is a docstring for
                the class.
        """
        assert isinstance(self._input, (ast.FunctionDef, ast.AsyncFunctionDef))
        docstring = ast.get_docstring(self._input)
        if docstring and docstring != "":
            style = self._get_parsing_style()
            parsed_docstring = None
            try:
                parsed_docstring = dp.parse(docstring, style=style)
            except dp.ParseError as e:
                raise ValueError(
                    self._get_error_location_msg()
                    + f"Failed to parse docstring using {style} style."
                ) from e

            return FunctionParser._get_docstring_data(parsed_docstring)
        else:
            return None

    def get_diffs(self) -> List[Diff]:
        """Gets the list of diffs for the node

        Returns:
            List[Diff]: List of Diffs for the node
        """
        if self._docstring_data is None:
            return [
                Diff(DiffReasons.NO_DOCSTRING, location=[], code_value=self._code_data)
            ]

        if self._docstring_data == self._code_data:
            return []

        assert self._documentable_type
        ignore_config = self._config.get_ignore_filters(self._documentable_type)
        diffs = self._get_arguments_list_diff(ignore_config)

        return_type_diff = self._get_type_diff(
            self._code_data.return_type,
            self._docstring_data.return_type,
            ["return_type"],
        )
        if return_type_diff:
            should_ignore = (
                ignore_config.get_ignore_return()
                and not self._config.fail_ignored_if_incorrect
            ) or (
                ignore_config.get_ignore_return()
                and return_type_diff.reason == DiffReasons.MISSING
            )
            if not should_ignore:
                print(f"reason:{return_type_diff.reason} lala")
                print(f"{ignore_config.get_ignore_return()}")
                diffs.append(return_type_diff)
        return diffs

    def _get_arguments_list_diff(self, ignore_config: IgnoreFilters) -> List[Diff]:
        assert self._docstring_data
        if (
            ignore_config.get_ignore_arguments()
            and len(self._docstring_data.arguments) == 0
        ) or (
            ignore_config.get_ignore_arguments()
            and not self._config.fail_ignored_if_incorrect
        ):
            return []
        matcher = SequenceMatcher()
        diffs: List[Diff] = []
        matcher.set_seqs(
            [a.name for a in self._docstring_data.arguments],
            [a.name for a in self._code_data.arguments],
        )
        docstring_shift = 0
        for _, i_doc, j_doc, i_code, j_code in matcher.get_opcodes():
            doc_args = self._docstring_data.arguments[i_doc:j_doc]
            code_args = self._code_data.arguments[i_code:j_code]
            for i, (code_arg, doc_arg) in enumerate(
                zip_longest(code_args, doc_args, fillvalue=None)
            ):
                index = i + i_doc + docstring_shift
                if not (code_arg and doc_arg):
                    if (
                        code_arg
                        and code_arg.name in ("*args", "**kwargs")
                        and ignore_config.ignore_args_and_kwargs
                    ):
                        continue
                    diffs.append(
                        Diff(
                            FunctionParser._get_element_diff_reason(code_arg, doc_arg),
                            location=["arguments", index],
                            code_value=code_arg,
                            docstring_value=doc_arg,
                        )
                    )
                else:
                    if code_arg.name != doc_arg.name:
                        diffs.append(
                            Diff(
                                DiffReasons.CONFLICTING_VALUE,
                                location=["arguments", index, "name"],
                                code_value=code_arg.name,
                                docstring_value=doc_arg.name,
                            )
                        )
                    type_diff = self._get_type_diff(
                        code_arg.type_hint,
                        doc_arg.type_hint,
                        ["arguments", index, "type_hint"],
                    )
                    if type_diff:
                        diffs.append(type_diff)
            shift = len(code_args) - len(doc_args)
            docstring_shift += shift if shift > 0 else 0
        return diffs

    def _get_type_diff(
        self,
        code_value: Optional[str],
        doc_value: Optional[str],
        location: List[Union[str, int]] = [],
    ) -> Optional[Diff]:
        assert self._documentable_type

        # We replace all new lines with spaces. It is fine to have multiline types.
        code_value = code_value.replace("\n", " ") if code_value else None
        doc_value = doc_value.replace("\n", " ") if doc_value else None
        # For the code value we strip quotes because those can be omitted in docstring
        # (and its fine)
        no_quotes_code_value = code_value.replace('"', "") if code_value else None
        if code_value != doc_value and no_quotes_code_value != doc_value:
            reason = FunctionParser._get_element_diff_reason(code_value, doc_value)
            ignore_config = self._config.get_ignore_filters(self._documentable_type)
            should_ignore = (
                ignore_config.get_ignore_typing()
                and not self._config.fail_ignored_if_incorrect
            ) or (ignore_config.get_ignore_typing() and reason == DiffReasons.MISSING)
            if not should_ignore:
                return Diff(
                    reason=reason,
                    location=location,
                    code_value=code_value,
                    docstring_value=doc_value,
                )

    @staticmethod
    def _get_element_diff_reason(code: Any, doc: Any):
        if doc is None:
            return DiffReasons.MISSING
        if code is None:
            return DiffReasons.EXTRA
        return DiffReasons.CONFLICTING_VALUE

    @classmethod
    def _get_parsing_style(cls) -> dp.DocstringStyle:
        cfg = load_config()
        return cls._DOCSTRING_STYLES[cfg.docstring_format]

    @staticmethod
    def _get_docstring_data(docstring: dp.Docstring) -> FunctionData:
        function_args = tuple(
            FunctionArgument(param.arg_name, param.type_name)
            for param in docstring.params
        )
        return_type = docstring.returns.type_name if docstring.returns else None
        return FunctionData(arguments=function_args, return_type=return_type)



"""
The core module defines the core functionalities in doXtrings execution pipeline.
The execution pipeline is:
    search: search for the files to be analyzed
    parse: parse the files found
    check: check the parsed documentables found
    report: report the findings of the analysis
"""
